/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */

/* (C) 2001 by Argonne National Laboratory.*/

/* MPI helper functions for long messages (send/recv/get_count with Aint counts) */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include "mpi.h"

typedef struct {
    MPI_Status mpistatus;
    MPI_Aint long_count;
    int MPE_LONG_SOURCE;
    int MPE_LONG_TAG;
} MPE_Long_status;

#define SHORT 100000000

int MPE_Long_send(void *buf, MPI_Aint count, MPI_Datatype datatype,
		  int dest, int tag, MPI_Comm comm);

int MPE_Long_recv(void *buf, MPI_Aint cound, MPI_Datatype datatype,
	      int source, int tag, MPI_Comm comm, MPE_Long_status *status);

int MPE_Long_get_count(MPE_Long_status status, MPI_Datatype datatype, MPI_Aint *count);

void printcomma(int);

int MPE_Long_send(void *buf, MPI_Aint count, MPI_Datatype datatype,
		  int dest, int tag, MPI_Comm comm)
{
    MPI_Aint numleft;
    int rc;

    if (datatype != MPI_BYTE) {
        printf("Only MPI_BYTE supported for long sends at this time\n");
        MPI_Abort(comm, 111);
    }

    numleft = count;
    while (numleft >= SHORT) {
        rc = MPI_Send(buf, SHORT, datatype, dest, tag, comm);
        buf     += SHORT;
        numleft -= SHORT;
    }
    rc = MPI_Send(buf, numleft, datatype, dest, tag, comm);

    return(rc);
}

int MPE_Long_recv(void *buf, MPI_Aint count, MPI_Datatype datatype,
	      int source, int tag, MPI_Comm comm, MPE_Long_status *status)
{
    int rc, recvd;
    MPI_Aint sofar;

    if (datatype != MPI_BYTE) {
        printf("Only MPI_BYTE supported for long receives at this time\n");
        MPI_Abort(comm, 111);
    }

    sofar = 0;
    while (sofar < count) {
        rc = MPI_Recv(buf, SHORT, datatype, source, tag, comm, &status->mpistatus);
        MPI_Get_count(&status->mpistatus, datatype, &recvd);
        sofar += recvd;
        if (sofar > count) { /* buffer too small; abort */
            printf("buffer too short, expected %ld, got %ld\n", count, sofar);
            MPI_Abort(comm, 222);
        }
        else if ((recvd < SHORT) || (sofar == count)) /* have read all of message */
            break;
        else {                  /* buffer not yet full and more expected, receive again; */
            buf += SHORT;
        }
    }

    status->long_count = sofar;
    status->MPE_LONG_SOURCE = (status->mpistatus).MPI_SOURCE;
    status->MPE_LONG_TAG    = (status->mpistatus).MPI_TAG;

    return(rc);
}


int MPE_Long_get_count(MPE_Long_status status, MPI_Datatype datatype, MPI_Aint *count)
{
    if (datatype != MPI_BYTE) {
        printf("Only MPI_BYTE supported for long Get_count at this time\n");
        MPI_Abort(MPI_COMM_WORLD, 111);
    }

    *count = status.long_count;

    return 0;
}

/* test program */

#define MSG_SIZE 10000000000

void printfcomma (long int n)
{
    if (n < 0) {
        printf ("-");
        printfcomma (-n);
        return;
    }
    if (n < 1000) {
        printf ("%ld", n);
        return;
    }
    printfcomma (n/1000);
    printf (",%03ld", n%1000);
}

int main(int argc, char *argv[])
{
    int size, rank;
    char *buf;
    MPE_Long_status status;
    MPI_Aint i, count, msg_size;

    MPI_Init(NULL, NULL);

    MPI_Comm_size(MPI_COMM_WORLD, &size);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);

    if (size < 2) {
        printf("This program needs at least two processes\n");
        return(0);
    }

    msg_size = MSG_SIZE;
    buf = malloc(msg_size);
    if (rank == 0) {
        double starttime, inittime, endtime;

        printf("message size is ");
        printfcomma(msg_size);
        printf("\n");
        starttime = MPI_Wtime();
        for (i = 0; i < msg_size; i++) buf[i] = '1';
        inittime = MPI_Wtime();
        printf("%d initialized buffer\n", rank);
        MPE_Long_send(buf, msg_size, MPI_BYTE, 1, 22, MPI_COMM_WORLD);
        endtime = MPI_Wtime();
        printf("rank 0 sent %ld in chunks of %d\n", msg_size, SHORT);
        printf("rank 0: inittime = %f, sendtime = %f\n",
               inittime - starttime, endtime - inittime);
    }
    else if (rank == 1) {
        double starttime, inittime, endtime;

        starttime = MPI_Wtime();
        for (i = 0; i < msg_size; i++) buf[i] = '2';
        printf("%d initialized buffer\n", rank);
        printf("%d before recv, buf = %c, %c, ..., %c, %c\n", rank,
               buf[0], buf[1], buf[msg_size-2], buf[msg_size-1]);
        inittime = MPI_Wtime();
        MPE_Long_recv(buf, msg_size, MPI_BYTE, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status);
        endtime = MPI_Wtime();
        MPE_Long_get_count(status, MPI_BYTE, &count);
        printf("rank %d received source = %d, tag = %d, count = %ld\n",
               rank, status.MPE_LONG_SOURCE, status.MPE_LONG_TAG, count);
        printf("%d after recv, buf = %c, %c, ..., %c, %c\n", rank, 
               buf[0], buf[1], buf[msg_size-2], buf[msg_size-1]);
        printf("rank 1: inittime = %f, recvtime = %f\n",
               inittime - starttime, endtime - inittime);
    }
    
    MPI_Finalize();
    return 0;
}
